﻿/*
 * Copyright 2011 Arne Vajhøj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;

using Mono.Cecil;
using Mono.Cecil.Cil;

namespace Vajhoej.StaticWeaver.Implementation
{
    /// <summary>
    /// Class AssemblyProcessor processes input and aspect assemmblies, generate glue assembly and creates
    /// modified versions of th einput assemblies.
    /// Note: this class uses Mono.Cecil.
    /// </summary>
    public class AssemblyProcessor
    {
        private delegate void Handler(ModuleDefinition mod, Instruction inst, Signature2Pointcut s2p, Pointcut2Advice p2a, MethodList ml, bool dbg);
        private static void Gen(ModuleDefinition mod, Instruction inst, Signature2Pointcut s2p, Pointcut2Advice p2a, MethodList ml, bool dbg)
        {
            Signature sig = new Signature(inst.Operand.ToString());
            IList<AdviceInfo> adv = AspectLoad.Find(sig, s2p, p2a);
            if(!sig.Internal && adv.Count > 0 && !sig.MethodName.StartsWith("."))
            {
                ml.Add(sig, adv);
                if(dbg)
                {
                    Console.WriteLine("Signature " + sig + " put on list");
                }
            }
        }
        private static void Chg(ModuleDefinition mod, Instruction inst, Signature2Pointcut s2p, Pointcut2Advice p2a, MethodList ml, bool dbg)
        {
            Signature sig = new Signature(inst.Operand.ToString());
            if(ml.Contains(sig))
            {
                Signature newsig = sig.Convert();
                if(ml.IsStatic(sig))
                {
                    inst.Operand = mod.Import(ml.GetType(newsig.TypeName).GetMethod(newsig.MethodName, newsig.ParameterDotnetTypes));
                }
                else
                {
                    if(inst.OpCode == OpCodes.Callvirt)
                    {
                        inst.OpCode = OpCodes.Call;
                    }
                    Type[] extraparm = new Type[1 + newsig.ParameterTypes.Count];
                    extraparm[0] = ml.GetType(sig.TypeName);
                    Array.Copy(newsig.ParameterDotnetTypes, 0, extraparm, 1, newsig.ParameterTypes.Count);
                    inst.Operand = mod.Import(ml.GetType(newsig.TypeName).GetMethod(newsig.MethodName, extraparm));
                }
                if(dbg)
                {
                    Console.WriteLine("Signature " + sig + " changed to " + newsig);
                }
            }
        }
        private static void Scan(AssemblyDefinition asm, Handler h, Signature2Pointcut s2p, Pointcut2Advice p2a, MethodList ml, bool dbg)
        {
            foreach(ModuleDefinition mod in asm.Modules)
            {
                if(mod.HasTypes)
                {
                    foreach(TypeDefinition typ in mod.Types)
                    {
                        if(typ.HasMethods && typ.CustomAttributes.Where(a => a.AttributeType.FullName == "Vajhoej.StaticWeaver.AspectAttribute").Count() == 0)
                        {
                            foreach(MethodDefinition met in typ.Methods)
                            {
                                if(dbg)
                                {
                                    Console.WriteLine("Scanning " + typ.FullName + "." + met.Name);
                                }
                                if(met.HasBody)
                                {
                                    foreach(Instruction inst in met.Body.Instructions)
                                    {
                                        if(inst.OpCode == OpCodes.Call || inst.OpCode == OpCodes.Callvirt)
                                        {
                                            h(mod, inst, s2p, p2a, ml, dbg);
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        private const string NEW_SUFFIX = "_yaaopf";
        private const string OLD_SUFFIX = "_backup";
        /// <summary>
        /// Weave aspects into assemblies.
        /// </summary>
        /// <param name="inputassemblies">List of names of input assemblies</param>
        /// <param name="glueassembly">Name of glue assembly</param>
        /// <param name="aspectassemblies">List of names of aspect assemblies</param>
        /// <param name="dbg">Debug flag (false =></param> no debug info, true => debug info written to Console.Out)</param>
        public static void Weave(IList<string> inputassemblies, string glueassembly, IList<string> aspectassemblies, bool dbg)
        {
            Signature2Pointcut s2p = new Signature2Pointcut();
            Pointcut2Advice p2a = new Pointcut2Advice();
            MethodList ml = new MethodList();
            foreach(string aa in aspectassemblies)
            {
                ml.SetAspectAssembly(aa);
                AspectLoad.Search(aa, s2p, p2a, dbg);
            }
            foreach(string ia in inputassemblies)
            {
                ml.SetInputAssembly(ia);
                AssemblyDefinition asm = AssemblyDefinition.ReadAssembly(ia);
                Scan(asm, Gen, s2p, p2a, ml, dbg);
            }
            CodeGen.Gen(ml, glueassembly, dbg);
            foreach(string ia in inputassemblies)
            {
                ml.SetInputAssembly(ia);
                AssemblyDefinition asm = AssemblyDefinition.ReadAssembly(ia);
                Scan(asm, Chg, s2p, p2a, ml, dbg);
                asm.Write(ia + NEW_SUFFIX);
            }
        }
    }
}
