﻿/*
 * Copyright 2011 Arne Vajhøj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

namespace Vajhoej.StaticWeaver.Implementation
{
    /// <summary>
    /// Class Signature contains information about a method signature.
    /// </summary>
    public class Signature
    {
        private const string TOP_NS = "Vajhoej.StaticWeaver.Glue.";
        private const string NS_PREFIX = "VSW_";
        private const string METHODNAME_PREFIX = "VSW_";
        private string rettyp;
        private string typnam;
        private string metnam;
        private IList<string> parmtyps;
        private static readonly Regex re = new Regex(@"^(\*|([\w]+(\.[\w]+)*(\[\])?)) (\*|[\w]+(\.[\w]+)*)::(\*|[\.\w]+)\((\*|[^\)]*)\)$", RegexOptions.Compiled);
        /// <summary>
        /// Create instance of Signature with the specified properties.
        /// </summary>
        /// <param name="rettyp">Return type</param>
        /// <param name="typnam">Type name</param>
        /// <param name="metnam">Method name</param>
        /// <param name="parmtyps">List of paramter types</param>
        public Signature(string rettyp, string typnam, string metnam, IList<string> parmtyps)
        {
            this.rettyp = rettyp;
            this.typnam = typnam;
            this.metnam = metnam;
            this.parmtyps = parmtyps;
        }
        /// <summary>
        /// Create instance of Signature with the specified properties.
        /// </summary>
        /// <param name="s">Text reprensentation of signature using CIL format (as used by Mono.Cecil)</param>
        public Signature(string s)
        {
            Match m = re.Match(s);
            if(m.Success)
            {
                rettyp = m.Groups[1].Value;
                typnam = m.Groups[5].Value;
                metnam = m.Groups[7].Value;
                parmtyps = new List<String>(m.Groups[8].Value.Split(",".ToCharArray(), StringSplitOptions.RemoveEmptyEntries));
            }
            else
            {
                throw new ArgumentException("Unsupported signature format: " + s);
            }
        }
        /// <summary>
        /// Return type.
        /// </summary>
        public string ReturnType
        {
            get { return rettyp; }
        }
        /// <summary>
        /// Type name.
        /// </summary>
        public string TypeName
        {
            get { return typnam; }
        }
        /// <summary>
        /// Method name.
        /// </summary>
        public string MethodName
        {
            get { return metnam; }
        }
        /// <summary>
        /// List of parameter types.
        /// </summary>
        public IList<string> ParameterTypes
        {
            get { return parmtyps; }
        }
        /// <summary>
        /// Parameter types as string.
        /// </summary>
        public string ParameterString
        {
            get { return parmtyps.Count == 0 ? "" : parmtyps.Aggregate((a,s) => a = a + "," + s); }
        }
        /// <summary>
        /// List of .NET parameter types (System.Type[]).
        /// </summary>
        public Type[] ParameterDotnetTypes
        {
            get
            {
                Type[] res = new Type[parmtyps.Count];
                for(int i = 0; i < res.Length; i++)
                {
                    res[i] = Type.GetType(parmtyps[i], true, false);
                }
                return res;
            }
        }
        /// <summary>
        /// Parameter types and arguments as string.
        /// </summary>
        public string ParameterStringWithArgs
        {
            get
            {
                StringBuilder sb = new StringBuilder();
                for(int i = 0; i < parmtyps.Count; i++)
                {
                    if(i > 0) sb.Append(",");
                    sb.Append(parmtyps[i]);
                    sb.Append(" ");
                    sb.Append("a" + (i + 1));
                }
                return sb.ToString();
            }
        }
        /// <summary>
        /// Arguments only as string.
        /// </summary>
        public string ArgsOnly
        {
            get
            {
                StringBuilder sb = new StringBuilder();
                for(int i = 0; i < parmtyps.Count; i++)
                {
                    if(i > 0) sb.Append(",");
                    sb.Append("a" + (i + 1));
                }
                return sb.ToString();
            }
        }
        /// <summary>
        /// Namespace part of type name.
        /// </summary>
        public string NameSpace
        {
            get { return typnam.Contains(".") ? typnam.Substring(0, typnam.Length - ClassName.Length - 1) : ""; }
        }
        /// <summary>
        /// Class name part of type name.
        /// </summary>
        public string ClassName 
        {
            get { return typnam.Split('.').Last(); }
        }
        /// <summary>
        /// Internal signature (part of StaticWeaver or glue assembly).
        /// </summary>
        public bool Internal
        {
            get { return typnam.StartsWith(TOP_NS) || metnam.StartsWith(METHODNAME_PREFIX); }
        }
        /// <summary>
        /// Convert to signature in glue assembly.
        /// </summary>
        /// <returns></returns>
        public Signature Convert()
        {
            return new Signature(rettyp, TOP_NS + NS_PREFIX + typnam.Replace(".", "." + NS_PREFIX), METHODNAME_PREFIX + metnam, parmtyps);
        }
        public override string ToString()
        {
            return string.Format("{0} {1}::{2}({3})", rettyp, typnam, metnam, ParameterString);
        }
        public override bool Equals(object obj)
        {
            Signature other = obj as Signature;
            if (other == null) return false;
            return this.ToString() == other.ToString();
        }
        public override int GetHashCode()
        {
            return ToString().GetHashCode();
        }
        
        public static bool operator ==(Signature lhs, Signature rhs)
        {
            if (ReferenceEquals(lhs, rhs)) return true;
            if (ReferenceEquals(lhs, null) || ReferenceEquals(rhs, null)) return false;
            return lhs.Equals(rhs);
        }
        public static bool operator !=(Signature lhs, Signature rhs)
        {
            return !(lhs == rhs);
        }
        private static bool WildcardMatch(string s1, string s2)
        {
            return (s1 == "*" || s2 == "*" || s1 == s2);
        }
        /// <summary>
        /// Wildcard match for signatures.
        /// Each of the signature components (return type, type name, method name and parameters) can
        /// be matched by a *.
        /// </summary>
        /// <param name="sig1">Signature</param>
        /// <param name="sig2">Signature</param>
        /// <returns>Match</returns>
        public static bool WildcardMatch(Signature sig1, Signature sig2)
        {
            return WildcardMatch(sig1.ReturnType, sig2.ReturnType)
                   && WildcardMatch(sig1.TypeName, sig2.TypeName)
                   && WildcardMatch(sig1.MethodName, sig2.MethodName)
                   && WildcardMatch(sig1.ParameterString, sig2.ParameterString);
        }
    }
}
